# -*- coding: utf-8 -*-
"""
Created on Wed Nov  1
Modified on Nov 17

@author: johana
Feature extractor's functions
"""
import numpy as np
import math
from skimage.feature import greycomatrix,greycoprops
from PIL import Image
import scipy.misc

ANGC=30
ANGP=90
DSIZE=256
CHANNEL=2

def samfeatures(sample,dsize):   
    mu=0
    s3=0
    ske=0
    s4=0
    kur=0
    var=0
    dest=0
    gra=0
    for i in range(dsize-1):
        mu=mu+sample[i]
    mu=mu/dsize
    for i in range(dsize-1):
        var=var+(sample[i]-mu)**2
        s3=s3+(sample[i]-mu)**3
        s4=s4+(sample[i]-mu)**4
    s3=s3/dsize
    s4=s4/dsize
    if dsize > 1:
        gra=sample[dsize-1]-sample[dsize-2]
        var=var/(dsize-1)
        dest=math.sqrt(var)
        ske=s3/dest**3
        kur=(s4/var**2)-3
    return np.array([max(sample),min(sample),mu,dest,ske,kur,gra])

def pdfeatures(image,dsize):
    tam=image.shape
    X=np.reshape(image,[tam[0]*tam[1]*tam[2],1])
    pdf,edges=np.histogram(X,bins=range(dsize),density=True)    
    mu=0
    s3=0
    s4=0
    var=0
    for i in range(dsize-1):
        mu=mu+i*pdf[i]
        s3=s3+pdf[i]*i**3
        s4=s4+pdf[i]*i**4

    for i in range(dsize-1):
        var=var+pdf[i]*(i-mu)**2
    dest=math.sqrt(var)
    ske=(s3-3*mu*var-mu**3)/dest**3
    kur=(s4-4*mu*s3+6*var*mu**2+3*mu**4)/dest**4-3
    return np.array([max(pdf),min(pdf),mu,dest,ske,kur])

def texfeaturesGLCM(data,dsize):
    glcmt=greycomatrix(data,[1],[0],dsize,normed=True)
    energy=greycoprops(glcmt, 'energy')
    cont=greycoprops(glcmt, 'contrast')
    homo=greycoprops(glcmt, 'homogeneity')
    corr=greycoprops(glcmt, 'correlation')
    return np.array([energy[0,0],cont[0,0],homo[0,0],corr[0,0]])

def texfeaturesSDH(imagen,dsize):
    image=imagen[:,:,CHANNEL]    
    tam=image.shape
    ims=image[:,0:tam[1]-2]+image[:,1:tam[1]-1]
    imd=abs(image[:,0:tam[1]-2]-image[:,1:tam[1]-1])
    ts=ims.shape
    td=imd.shape
    imsv=np.reshape(ims,[ts[0]*ts[1],1])
    imdv=np.reshape(imd,[td[0]*td[1],1])
    pdfs,edges=np.histogram(imsv,bins=range(2*dsize),density=True) 
    pdfd,edges=np.histogram(imdv,bins=range(dsize),density=True) 
    ened=pdfd*np.transpose(pdfd)
    ened=ened.sum(axis=0)
    j2=np.power(range(dsize-1),2) 
    cont=pdfd*np.transpose(j2)
    cont=cont.sum(axis=0)
    one=np.ones(dsize-1)    
    cta=1/(one+j2)
    homo=pdfd*np.transpose(cta)
    homo=homo.sum(axis=0)    
    enes=pdfs*np.transpose(pdfs)
    enes=enes.sum(axis=0)
    ies=range(2*dsize-1)
    one=np.ones(2*dsize-1)
    mu=pdfs*np.transpose(ies)
    mu=mu.sum(axis=0)
    cor1=pdfs*np.transpose(np.power((ies-one*mu),2))
    cor1=cor1.sum(axis=0)
    energy=ened*enes
    corr=1/2*(cor1-cont)
    return np.array([energy,cont,homo,corr])  

def imagefeature(image,case=0):
    features=([])
    features=pdfeatures(image,DSIZE)
    if case==1:
        features=np.append(features,texfeaturesSDH(image,DSIZE))
    else:
        features=np.append(features,texfeaturesGLCM(image[:,:,CHANNEL],DSIZE))
    return features
    
def load_image_array( infilename ) :
    img = Image.open( infilename )
    img.load()
    data = np.asarray( img, dtype="int32" )
    return data

def save_image( data, outfilename ) :
    scipy.misc.imsave(outfilename,data)