"""
Created on Wed Oct 18 2017
Modified on July 20 2020

@author: johana
"""

#**** Libraries
from osbrain import run_agent
from osbrain import run_nameserver
from osbrain import Agent
import numpy as np

# library NEAT
import fileinput
import neat
import os
import gzip
import pickle

# Library  GFS
from GFS_Agents import *

#------------------------------ Global variables
DAY = 31  # Sample DAY
sDAY = 32 # Training DAY
NDATA = 3 
AGENTS = 5
STATES = NDATA*AGENTS
MFEATURES =10  # Maximum number of features per agent
WAITIME = 5
EMPTYS = np.zeros(STATES)
DMS_CUN = 'ALL' # 'NEAT' or 'GFS' or 'MEAN' or 'MEDIAN' or 'MAX' or 'VOTE' or 'ALL'
NFEA = 7  # Number of features per agent

# Physical address
TRP='ipc'
add1='127.0.0.1:11100'
add2='127.0.0.1:11200'
add3='127.0.0.1:11300'
add4='127.0.0.1:11400'
add5='127.0.0.1:11500'
addc='127.0.0.1:11000'

#------------------------------ Class
        
class ModelAgent(Agent):
    """
    This function creates and defines the agent, the communication attibute,
    the internal variables and the internal variables call methods. The methods
    cannot be call from the main function. Only the functions part of the class 
    can be called.
    """
    def on_init(self):    
        # Internal variables
        self.set_attr(agent_states=np.zeros(STATES),features_sensor=np.zeros(MFEATURES),restop = np.zeros(2))
        self.set_attr(features=0.0, control=0, inputs=0, idx=0)
        # Internal methods - private functions
        # Methods of variables        
        self.set_method(set_array_states, set_array_features, set_control, set_features, set_inputs, set_idx, set_rs,reset_rs)
        self.set_method(get_array_states, get_array_features, get_control, get_features, get_inputs, get_idx, get_rs)
        # Operation methods
        self.set_method(reset_arr_fea)

    def phy_address(self,num,padr):
        # Communication 
        self.bind('PUB',transport= TRP,addr= padr, alias='main')
        self.set_idx((num-1)*NDATA)
    
    def rscanal(self,padr):
        self.bind('PUB',transport= TRP,addr= padr, alias='control')

    def softreset(self):
        self.save_info()
        self.set_array_states(EMPTYS, 0)
        self.reset_arr_fea()
    
    def __del__(self):
        self.kill()
    
    
    """
    These functions are executed when a message is received. The source of 
    the message selects the correct function. The messages are defined per each
    sensor, however, there is a message function per agent that it is not going
    to be applied.
    
    Message: Next-Position, Local Uncertain-measure, Actual-Position 
    """
    def message_S1(self, message):
        #self.log_info(message)
        self.set_array_states(message,0)
        
    def message_S2(self, message):
        #self.log_info(message)        
        self.set_array_states(message,NDATA)

    def message_S3(self, message):
        #self.log_info(message)
        self.set_array_states(message,2*NDATA)
        
    def message_S4(self, message):
        #self.log_info(message)
        self.set_array_states(message,3*NDATA)
        
    def message_S5(self, message):
        #self.log_info(message)
        self.set_array_states(message,4*NDATA)
    
    def message_RS(self, message):
        # RESET STOP
        self.log_info(message)
        if  message[0] == 1 :
            self.softreset()
            if (self.name  != 'TEM'):
                message[0] == 0
        self.set_rs(message)
        
        
    """
    This function sends a message that are listen to the subscribers. The 
    message has three parts, the position 0 contains the next angle of the module
    and the position 1 has the uncertain measure and the position 2 has the actual
    angle.
    """
    def info(self,message_out): # Next best position, Local Uncertain, Actual posicion
        idx=self.get_idx()
        #self.log_info(message_out)
        self.set_array_states(message_out,idx)
        self.send('main', message_out)
    
    def menrs(self,message):    # Reset - Stop
        self.log_info(message)
        self.send('control', message)
        if message[0] == 1 :
            self.softreset()
            message[0] = 0
    
    def save_info(self):
        ind = self.get_control()
        cluster = np.append(self.features_sensor,self.agent_states)
        intro = 'Data/'+ self.name + '_pos_'
        np.savetxt(intro + str(ind) +'.csv',cluster,delimiter=',')
        ind=ind+1
        self.set_control(ind)
      
    """
    This function receive an external fuction and execute it, to obtain a new 
    sensor acquisition. It returns the value of the sensor.
    """
    def sensor_read(self,funct,*args):
        a=self.execute_as_function(funct,*args)     
        return a
        
    """
    This function receive an external function which extract the features
    from the data sensor, and store them in a variable. The features have a
    maximum value (MFEATURES) and could be differ among agents so the 
    number of features is a variable that is set in this function.
    """
    def extract_features(self,funct,*args):
        a=self.execute_as_function(funct,*args)
        self.set_features(len(a))
        self.set_inputs(int(self.get_features()) + int(STATES))
        #self.log_info(self.get_features())
        for x in range(self.get_features()):
            self.set_array_features(a[x],x)
    
    def IS_Inputs(self):
        cluster = np.append(self.features_sensor[0:self.features],self.agent_states)
        return self.features_sensor[0:self.features],cluster,self.get_inputs()
    
    def nbp_read(self):
        idx=self.get_idx()
        return self.get_array_states(idx)
    
    def getdatafeature(self,num):
        return self.get_array_features(num)
    
    def setdatafeature(self,data,num):
        self.set_array_features(data, num)
    
    def readstop(self):
        return self.get_rs(1)
        
    def readreset(self):
        a = self.get_rs(0)
        self.reset_rs(0)
        return a
        
        

#------------------------------ Other functions which are part of the internal process

# Variable functions
"""
These functions are called as methods of the agent model. They affect the 
internal variable of the agent. They apply for every agent.
"""
def set_array_states(self, value, num):
    tam=len(value)
    for i in range(tam):
        self.agent_states[num+i] = value[i]

def get_array_states(self, num):
    return self.agent_states[num]

def set_array_features(self, value, num):
    self.features_sensor[num] = value

def reset_arr_fea(self):
    for i in range(MFEATURES):
        self.set_array_features(0,i)
    
def get_array_features(self,  num):
    return self.features_sensor[num]
    
def set_control(self, value):
    self.control = value

def get_control(self):
    return self.control
   
def set_features(self, value):
    self.features = value

def get_features(self):
    return self.features

def set_inputs(self, value):
    self.inputs = value

def get_inputs(self):
    return self.inputs

def set_idx(self, value):
    self.idx = value

def get_idx(self):
    return self.idx   

def set_rs(self, value):
    tam=len(value)
    for i in range(tam):
        self.restop[i] = value[i]

def get_rs(self,num):
    return self.restop[num]

def reset_rs(self,num):
    self.restop[num] = 0

#------------------------------ Functions Intelligent System
"""
Selection function
"""
def a_DMS(tsensor,net_file, netA_file, inputnetfea, inputComplete, tam, idx,fsinB,fsouB,frulB):
    # Internal NET - Apply in each case
    win_net = restore_net(net_file)
    fea = ()
    tamf = inputnetfea.shape
    for j in range (tamf[0]):
        fea = fea + (inputnetfea[j],)
    uncl = win_net.activate(fea)
    
    # Features base
    base = NFEA

    # Coordinated Uncertain
        
    if DMS_CUN == 'NEAT':
        pos=()
        for j in range (tam):
            pos = pos + (inputComplete[j],)
        win_netA = restore_net(netA_file)
        uncc = win_netA.activate(pos)
    elif DMS_CUN == 'GFS':
        dati = inputdata(tsensor,inputComplete)
        uncct = FuzzyControlResult(fsinB,fsouB,frulB,dati,'UNC',tsensor)
        uncc = np.array([uncct])
    elif DMS_CUN == 'MEAN' or DMS_CUN == 'MEDIAN' or DMS_CUN == 'MAX' or DMS_CUN == 'VOTE':
        results = np.array([inputComplete[base+1],inputComplete[base+4],inputComplete[base+7],inputComplete[base+10],inputComplete[base+13]])
        if DMS_CUN == 'MEAN':
            uncc = np.array([np.mean(results)])
        elif DMS_CUN == 'MEDIAN':
            uncc = np.array([np.median(results)])
        elif DMS_CUN == 'MAX':
            uncc = np.array([np.amax(results)])
        else:
            vote = 0
            for k in range(AGENTS):
                if results[k] > 0.5:
                    vote = vote + 1
            if vote > 2:
                uncc = np.array([1])
            else:
                uncc = np.array([0])
    elif DMS_CUN == 'ALL':
        uncc = np.zeros([6])
       
        # NEAT
        pos=()
        for j in range (tam):
            pos = pos + (inputComplete[j],)
        win_netA = restore_net(netA_file)
        unct = win_netA.activate(pos)
        uncc[0]= unct[0]
        
        # GFS
        dati = inputdata(tsensor,inputComplete)
        uncct = FuzzyControlResult(fsinB,fsouB,frulB,dati,'UNC',tsensor)
        unct = np.array([uncct])
        uncc[1]= unct[0]
        
        # STADISTIC
        results = np.array([inputComplete[base+1],inputComplete[base+4],inputComplete[base+7],inputComplete[base+10],inputComplete[base+13]])
        unct = np.array([np.mean(results)])
        uncc[2]= unct[0]
        unct = np.array([np.median(results)])
        uncc[3]= unct[0]
        unct = np.array([np.amax(results)])
        uncc[4]= unct[0]
        vote = 0
        for k in range(AGENTS):
            if results[k] > 0.5:
                vote = vote + 1
        if vote > 2:
            unct = np.array([1])
        else:
            unct = np.array([0])
        uncc[5]= unct[0]
    else: 
        print('Error in Decision-Making Uncertain, selection not avalible')

    return uncl,uncc
    
"""
These functions are part of the intelligent system. Neural Networks NEAT
"""

def restore_net(filename):
    with gzip.open(filename) as f:
        return pickle.load(f)

"""
These function is for GFS Genetic Fuzzy System
"""

def a_DMS_FL_Rules(tsensor):
    # Files fuzzy
    
    Data_file ='FS/'+tsensor+'_UNC_GFS_genome.csv'
    indvUNC = np.genfromtxt(Data_file,delimiter=',')
    Data_file ='FS/'+tsensor+'_UNC_GFS_enable.csv'
    erulUNC = np.genfromtxt(Data_file,delimiter=',')
    
    # Fuzzy Control
    fsinB,fsouB,frulB = SystemFuzzySets(indvUNC,erulUNC)
    return fsinB,fsouB,frulB
