"""
Created on Tue Dec 12 2017
Modified on Feb 2 2018

@author: johana
Genetic Fuzzy System - Library
"""
# Libraries:
import random
import numpy as np
import skfuzzy as fuzz
import time


class GFS():
    # Genetic Algorithm
    cxrate, mtrate = 0.7, 0.2
    numinv = 50 #50 30
    vmaxfit = 0.1
    ngene = 80 #80 #100
    elitind = int(numinv*5/100)
    muterat = 0.2
    dshow = 5
    
    # Fuzzy logic
    nbs = 4  # 5: 32, 4: 16 
    nbv = 2  # 3: 8, 2: 4 : Input bits
    nbo = 3  # 4: 16, 3: 8 : output bits
    cin = 2  # 6:8 Cones.....2:4 Cones
    cot = 1
    cno = 6  # 14:16 Cones, 6: 8 Cones
    x_axis = np.arange(0, 2**nbs, 1)    
    
    def __init__(self,inputs,db):
        lastcol = 0
        self.db = db
        
        if db == 'C1':
            lastcol = 60
        elif db == 'C2':
            lastcol = 44
        elif db == 'C3':
            lastcol = 30
        else:
            lastcol = 20
            
        self.inp = inputs 
        
        # Train and evaluation data
        agent = 1        
        data_file='../NEAT/Result/'+str(inputs)+str(agent)+db+'_TD.csv'
        data_tem = np.genfromtxt(data_file,delimiter=',')
        self.samp = len(data_tem)
        data_total1 = np.zeros([self.samp,inputs + 1])
        data_total1[:,agent-1] = data_tem
        
        data_file='../NEAT/Result/'+str(inputs)+str(agent)+db+'_ED.csv'
        data_tem = np.genfromtxt(data_file,delimiter=',')
        self.same = len(data_tem)
        data_total2 = np.zeros([self.same,inputs + 1])
        data_total2[:,agent-1] = data_tem
        
        while (agent<inputs):
            agent = agent+1
            
            data_file='../NEAT/Result/'+str(inputs)+str(agent)+db+'_TD.csv'
            data_total1[:,agent-1] = np.genfromtxt(data_file,delimiter=',')
            
            data_file='../NEAT/Result/'+str(inputs)+str(agent)+db+'_ED.csv'
            data_total2[:,agent-1] = np.genfromtxt(data_file,delimiter=',')
        
        data_file='../DB/'+db+'_T.csv'
        data_tem = np.genfromtxt(data_file,delimiter=',')
        data_total1[:,inputs] = data_tem[:,lastcol]
        self.data_train = data_total1
        
        data_file='../DB/'+db+'_E.csv'
        data_tem = np.genfromtxt(data_file,delimiter=',')
        data_total2[:,inputs] = data_tem[:,lastcol]
        self.data_eval = data_total2
        

    # GENETIC ALGORITHM FUNCTIONS:
    
    # Individual set
    def indvSET(self,con,bit):
        lmax = 2**bit
        paso = int(lmax/con-1)    
        com = ([]) 
        for i in range(con):
            com = np.append(com, dec_to_bin(paso*(i+1),bit))
        return com
    
    def setSET(self,con,bit,nset):
        sets = ([])
        oset = self.indvSET(con,bit)    
        for i in range(nset):
            sets = np.append(sets,oset)
        return sets
    
    def sets(self,coni,cono,bits,nsi,nso):
        indv = ([])
        # Sets
        indv =  np.append(indv,self.setSET(coni,bits,nsi))
        indv =  np.append(indv,self.setSET(cono,bits,nso))
        nset = len(indv)
        return indv,nset
    
    # Individual rules
    def indvRules(self,lin,bit,inp):
        tem = np.zeros([self.inp])    
        result = self.rule(inp,lin,tem,([]))     
        tam = len(result)
        resb=([])
        for i in range(tam):
            tem = dec_to_bin(result[i],bit)
            resb = np.append(resb,tem)
        tam = len(resb)
        resb = np.resize(resb,[int(tam/(self.inp*bit)),(self.inp*bit)])
        return resb
    
    def rule(self,pos,num,tem,result):
        if pos>0:
            npos=pos-1
            for i in range (num):
                tem[pos]=i
                result = self.rule(npos,num,tem,result)
        else:
            for j in range (num):
                tem[pos]=j
                #print(tem)
                result = np.append(result,tem)
        return result
    
    # Individuo
    def indv (self,indv,nset,rules):
        # Rules output
        nrules = len (rules)
        for i in range (nrules):
            indv = np.append(indv,rules[i,:])
            a = random.randint(1, self.cno+1)
            ab = dec_to_bin(a,self.nbo)
            indv = np.append(indv,ab)
        ntot = len(indv)
        nrules = ntot-nset
        return(indv,nset,nrules)
    
    # Generate initial population
    def initpop(self,numinv):
        inicio = time.time()
        rules = self.indvRules(self.cin+2,self.nbv,self.inp-1)
        fin = time.time()
        print('rules '+str(fin-inicio)+' s')
        inicio = fin
        setst,nset = self.sets(self.cin,self.cno,self.nbs,self.inp,self.cot)
        fin = time.time()
        print('sets '+str(fin-inicio)+' s')
        inicio = fin
        pop=([])
        for i in range(numinv):
            arin,ns,nr = self.indv (setst,nset,rules)
            pop = np.append(pop,arin)
        tam = ns+nr
        fin = time.time()
        print('pop '+str(fin-inicio)+' s')
        print([numinv,nr])
        return np.resize(pop,[numinv,tam]),np.random.randint(2, size=[numinv,nr]),tam,nr
        #np.random.randint(2, size=[numinv,nr])
        #np.ones([numinv,nr])
        
    # Reproduction operators
    # Cross one points
    def crossO(self,ind1,ind2):
        size = len(ind1)
        cxpoint = random.randint(1, size)
        ind1[0:cxpoint], ind2[0:cxpoint] \
            = ind2[0:cxpoint].copy(), ind1[0:cxpoint].copy()      
        return ind1, ind2
    
    # Cross two points
    def crossT(self,ind1,ind2):
        size = len(ind1)
        cxpoint1 = random.randint(1, size)
        cxpoint2 = random.randint(1, size - 1)
        if cxpoint2 >= cxpoint1:
            cxpoint2 += 1
        else: # Swap the two cx points
            cxpoint1, cxpoint2 = cxpoint2, cxpoint1
    
        ind1[cxpoint1:cxpoint2], ind2[cxpoint1:cxpoint2] \
            = ind2[cxpoint1:cxpoint2].copy(), ind1[cxpoint1:cxpoint2].copy()      
        return ind1, ind2
    
    # Enable/Disable rules
    def EDRules(self,inv,edr):
        size = len(inv)
        for i in range(size):
            if random.random() < edr:
                inv[i]=not(inv[i])
        return inv
    
    # Displacement set
    def Dis_set(self,inv,sets,bit):
        a = random.randint(1, sets-1)
        ini = a*bit
        val = inv[ini:ini+bit]
        cen = bin_to_dec(val,bit)
        if random.random() < 0.5:
            cen=cen+1
        else:
            cen=cen-1
        val = dec_to_bin(cen,bit)
        inv[ini:ini+bit] = val
        return inv
    
    # Mutation, multi bit 
    def mutation(self,inv,mutr):
        size = len(inv)
        for i in range(size):
            if random.random() < mutr:
                inv[i]=not(inv[i])
        return inv
    
    # Select the best individuals
    def selbest(self,pop,ena,fitv,nbest):
        npop = np.copy(pop)
        nena = np.copy(ena)
        nfit = np.copy(fitv)
        sel = ([])
        enr = ([])
        val = np.zeros(nbest)
        for i in range(nbest):
            flag = True 
            print (len(nfit))
            if len(nfit) > 0:
                indd = np.argmax(nfit) 
                bestf = max(nfit)
                val[i] = bestf
                ind = nfit == bestf
                sel = np.append(sel,npop[indd])
                enr = np.append(sel,nena[indd])
                nfit = nfit[np.logical_not(ind)]
                npop = npop[np.logical_not(ind)]
                nena = nena[np.logical_not(ind)]
            else:
                flag = False
        return sel, enr, val, flag
    
    # Indicates the worst individuals
    def retworst(self,fitv,nbest):
        nfit = np.copy(fitv)
        indd = np.zeros(nbest)
        for i in range(nbest):
            a = np.argmin(nfit)
            indd[i] = a        
            worst = min(nfit)
            nfit[a] = 1000
        return indd
    
    # FUZZY FUNCTIONS:
    
    # Evaluate individual
    def ffitness(self,indv,erul):
        # Some useful sizes
        tam = len(indv)
        ten = len(erul)
        tsei = self.nbs*self.inp*self.cin
        tseo = self.nbs*self.cot*self.cno
        tcru = self.inp + self.cot
        tstt = tsei+tseo    
        trul = tam-tstt
        
        # Dividing the information among sets and rules
        dsei = indv[0:tsei]  
        dseo = indv[tsei:tstt]
        drul = indv[tstt:tam]
        a = len(drul)
        col = self.inp*self.nbv + self.cot*self.nbo
        row = int(a/col)    
        drul = np.resize(drul,[row,col])    
        
        # Transforming bin to dec
        sind = abin_to_adec(dsei,self.nbs)
        sopd = abin_to_adec(dseo,self.nbs)
        srud = ([])
        for k in range(row):
            auxi = abin_to_adec(drul[k,0:self.inp*self.nbv],self.nbv)
            auxo = abin_to_adec(drul[k,self.inp*self.nbv:col],self.nbo)
            srud = np.append(srud,auxi)
            srud = np.append(srud,auxo)
        
        # Ordered in an array
        sind = np.resize(sind,[self.inp,self.cin])
        sopd = np.resize(sopd,[self.cot,self.cno])
        srud = np.resize(srud,[ten,tcru])
        
        rr = 0    
        frul = ([])
        for i in range (max(self.inp,self.cot,ten)):
            if i < self.inp:
                sind[i,:] = np.sort(sind[i,:])
            if i < self.cot:
                sopd[i,:] = np.sort(sopd[i,:])
            if i < ten:
                if erul[i] == 1:
                    frul = np.append(frul,srud[i,:])
                    rr = rr+1
        frul = np.resize(frul,[rr,tcru])
        
        # Create fuzzy membership functions
        fsin = self.fuzzyset(sind,case = 0)
        fsou = self.fuzzyset(sopd,case = 1)
        
        # Data train
        fit = 0
        resfl = np.zeros(self.samp)
        for s in range(self.samp):
            dati = self.Tsample(s)
            dato = self.data_train[s,self.inp]
                
            # Fuzzyfication
            lini = self.appfuzzysample(dati,fsin,self.cin + 2)
            # Applying Fuzzy rules
            ouli = self.appfuzzyrules(lini,frul)
            lino = self.outfuzzrule(frul,ouli,fsou[0,:,:])
            outfuz = self.fuzzaggregate(lino)
            # Defuzzyfication
            fout = self.defuzzout(outfuz)
            resfl[s] = fout
            # fit evaluation
            delta = (fout - dato)**2
            fit = fit - delta
        
        # Data Eval
        fite = 0
        evafl = np.zeros(self.same)
        for s in range(self.same):
            dati = self.Esample(s)
            dato = self.data_eval[s,self.inp]
                
            # Fuzzyfication
            lini = self.appfuzzysample(dati,fsin,self.cin + 2)
            # Applying Fuzzy rules
            ouli = self.appfuzzyrules(lini,frul)
            lino = self.outfuzzrule(frul,ouli,fsou[0,:,:])
            outfuz = self.fuzzaggregate(lino)
            # Defuzzyfication
            fout = self.defuzzout(outfuz)
            evafl[s] = fout
            # fit evaluation
            delta = (fout - dato)**2
            fite = fite - delta
        
        return fit, fite,resfl, evafl
    
    def fuzzyset(self,crisp,case = 0):
        az = 2**self.nbs
        nsetsi = self.cin + 2
        nsetso = self.cno + 2
        delta = az/((self.cno+1)*2)
        if case == 0:
            result= np.zeros([self.inp,nsetsi,az])
            for j in range(self.inp):
                center = 0
                for k in range(nsetsi):
                    if k == 0:
                        result[j,k,:] = fuzz.trimf(self.x_axis, [0, center, crisp[j,k]])
                    elif 0 < k < nsetsi-2:
                        result[j,k,:] = fuzz.trimf(self.x_axis, [center, crisp[j,k-1], crisp[j,k]])
                        center=crisp[j,k-1]
                    elif k == nsetsi-2:
                        result[j,k,:] = fuzz.trimf(self.x_axis, [center, crisp[j,k-1], az-1])
                        center=crisp[j,k-1]
                    else:
                        result[j,k,:] =fuzz.trimf(self.x_axis, [center, az-1, az-1])
        else:
            result= np.zeros([self.cot,nsetso,az])
            for j in range(self.cot):
                center = 0
                for k in range(nsetso):
                    if k == 0:
                        result[j,k,:] = fuzz.trimf(self.x_axis, [0, center, delta])
                    elif 0 < k < nsetso-1:
                        center = crisp[j,k-1]
                        result[j,k,:] = fuzz.trimf(self.x_axis, [center-delta, center, center+delta])
                    else:
                        center = az-1
                        result[j,k,:] =fuzz.trimf(self.x_axis, [center-delta, center, center])
        return result
    
    def aplifuzzyset(self,dat,mmf):
        return fuzz.interp_membership(self.x_axis, mmf, dat)
    
    def evalfuzzysets(self,dat,setmf,nset):
        result = np.zeros(nset)    
        for i in range(nset):
            result[i] = self.aplifuzzyset(dat,setmf[i,:])
        return result
    
    def appfuzzysample(self,sample,sett,nset):
        result = np.zeros([nset,self.inp]) 
        for i in range (self.inp):
            result[:,i] = self.evalfuzzysets(sample[i],sett[i,:,:],nset)
        return result
    
    def appfuzzyrules(self,vlin,rules):
        nrul,var=rules.shape
        resulti = np.zeros(nrul)
        for i in range(nrul):
            resulti[i] = self.infuzzrule(vlin,rules[i,:])
        return resulti
            
    def infuzzrule(self,vlin,rule):
        rulet = 0
        for j in range(self.inp):
            ind = int(rule[j]) 
            rulet = np.fmax(rulet, vlin[ind,j])
        return rulet
    
    def outfuzzrule(self,rules,rin,sout):
        pout = 2**self.nbo
        a,b = sout.shape
        retout = np.zeros([pout,b])
        intout = np.zeros(pout)
        for j in range(pout):
            inds = rules[:,self.inp] == j# Restricted to one input       
            tmo = np.full_like(inds,1)
            tmot = tmo[inds]
            l = len(tmot)
            if l > 1:
                tempi = rin[inds]
                nrto = len(tempi)
                if nrto > 1:
                    tempo = 0
                    for k in range (nrto):
                        tempo = np.fmax(tempo,tempi[k])
                    intout[j] = tempo
                else:
                    intout[j] = tempi
        for k in range(pout):
            retout[k,:] = np.fmin(intout[k],sout[k,:])
        return retout
    
    def fuzzaggregate(self,seto):
        opt,lon = seto.shape
        ruleo = np.zeros(lon)
        for j in range(opt):
            ruleo = np.fmax(ruleo,seto[j,:])
        return ruleo
    
    def defuzzout(self,outfuz):
        outnor = fuzz.defuzz(self.x_axis, outfuz, 'centroid')
        outf = outnor/(2**self.nbs-1)
        return outf
        
    # Return sample:
    def Tsample(self,x):
        tem = self.data_train[x,:]
        result = np.zeros(self.inp) 
        for k in range(self.inp):
            result[k] = (2**self.nbs-1)*tem[k]
        return result
    
    def Esample(self,x):
        tem = self.data_eval[x,:]
        result = np.zeros(self.inp) 
        for k in range(self.inp):
            result[k] = (2**self.nbs-1)*tem[k]
        return result
    
    def run(self):
        inicio = time.time()
        popd,pope,tamgen,nrules = self.initpop(self.numinv)
        fin = time.time()
        print('Initial Population at '+str(fin-inicio)+' s')
        sets = self.inp*self.cin + self.cot*self.cno
        fit = np.zeros([self.numinv,2])
        
        print("Start of evolution "+ self.db + ' # inputs '+ str(self.inp))
        inicio = time.time()
        bfit = -1000000000
        
        # Evaluation of the initial population    
        for k in range(self.numinv):
            fit[k,0],fiteva,restrain,reseval = self.ffitness(popd[k,:],pope[k,:])
            fit[k,1] = 1
            if k % self.dshow == 0:
                fin = time.time()
                delta = fin-inicio
                inicio = fin
                print('Eval individuo '+str(k)+' of '+str(self.numinv)+' in '+str(delta)+ ' s')
                
            if fit[k,0] > bfit:
                bfit = fit[k,0]
                bfite = fiteva
                binv = popd[k,:]
                beru = pope[k,:]
                data_result = restrain
                data_evaluation = reseval
                
        # Variable keeping track of the number of generations
        g = 0
        data = ([]) 
        
        # Stadisticts of the evolution
        mean = sum(fit[:,0]) / self.numinv
        sum2 = sum (x*x for x in fit[:,0])
        std = abs(sum2 / self.numinv - mean**2)**0.5
        
        tem = np.zeros(8)
        tem[0] = g
        tem[1] = self.numinv
        tem[2] = min(fit[:,0])
        tem[3] = max(fit[:,0])
        tem[4] = mean
        tem[5] = std
        tem[6] = bfite
        tem[7] = self.samp
        data = np.append(data,tem)
        
        # Begin the evolution
        fitt = np.zeros([self.numinv,2])
        while max(fit[:,0]) < self.vmaxfit and g < self.ngene:
            # A new generation
            g = g + 1
            mss = '------- Generation '+str(int(g))+' of '+str(self.ngene)+' best fit '+ str(bfit)
            print(mss)
    
            # Selection of the best individuals and removal of the worst 
            if self.elitind > 0 :                      
                offspring, offena, tvfit, isok = self.selbest(popd,pope,fit[:,0],self.elitind)
                if isok:
                    ind = self.retworst(fit[:,0],self.elitind)
                    ind = np.sort(ind)
                    cpop = ([])
                    crpo = ([])
                    ori = 0
                    for n in range(self.elitind):
                        fitt[n,0] = tvfit[n]
                        fitt[n,1] = 1
                        if not (n==self.elitind-1):
                            a = int(ind[n])
                            cpop = np.append (cpop,popd[ori:a,:])
                            crpo = np.append (crpo,pope[ori:a,:])
                            ori = int(ind[n]+1)
                        else:
                            a = int(ind[n])
                            cpop = np.append (cpop,popd[ori:a,:])
                            crpo = np.append (crpo,pope[ori:a,:])
                            ori = int(ind[n]+1)
                            cpop = np.append (cpop,popd[ori:self.numinv,:])
                            crpo = np.append (crpo,pope[ori:self.numinv,:])
                        
                    cpop = np.reshape(cpop, [int(self.numinv-self.elitind),tamgen])
                    crpo = np.reshape(crpo, [int(self.numinv-self.elitind),nrules])
                    k = self.elitind
                    family = int((self.numinv-self.elitind)/2)
                else:
                    print('Generation without best survival')
                    offspring = ([])
                    offena = ([])
                    cpop = np.copy(popd)
                    crpo = np.copy(pope)
                    k = 0
                    family = int(self.numinv/2)
            else:
                offspring = ([])
                offena = ([])
                cpop = np.copy(popd)
                crpo = np.copy(pope)
                k = self.elitind
                family = int(self.numinv/2)
            
            # Apply crossover and mutation on the offspring
                    
            for inin in range (family):        
                if random.random() < self.cxrate:
                    rat = random.random() 
                    if rat < 0.4:
                        if random.random() < 0.5:
                            temp1, temp2 = self.crossT(cpop[2*inin,:], cpop[2*inin+1,:])
                            offspring = np.append(offspring,[temp1,temp2])
                            offena = np.append(offena,[crpo[2*inin,:], crpo[2*inin+1,:]])
                        else:
                            temp1, temp2 = self.crossO(cpop[2*inin,:], cpop[2*inin+1,:])
                            offspring = np.append(offspring,[temp1,temp2])
                            offena = np.append(offena,[crpo[2*inin,:], crpo[2*inin+1,:]])
                    elif 0.4 <= rat < 0.7:
                        temp1 = self.EDRules(crpo[2*inin,:],self.muterat)
                        temp2 = self.EDRules(crpo[2*inin+1,:],self.muterat)
                        offena = np.append(offena,[temp1,temp2])
                        offspring = np.append(offspring,[cpop[2*inin,:],cpop[2*inin+1,:]])
                    else:
                        temp1 = self.Dis_set(cpop[2*inin,:],sets,self.nbs)
                        temp2 = self.Dis_set(cpop[2*inin+1,:],sets,self.nbs)
                        offspring = np.append(offspring,[temp1,temp2])
                        offena = np.append(offena,[crpo[2*inin,:], crpo[2*inin+1,:]])
                else:
                    offspring = np.append(offspring,[cpop[2*inin,:],cpop[2*inin+1,:]])
                    offena = np.append(offena,[crpo[2*inin,:], crpo[2*inin+1,:]])
                    fitt[k ,0] = fit[2*inin,0]
                    fitt[k+1,0] = fit[2*inin+1,0]
                    fitt[k,1] = 1
                    fitt[k+1,1] = 1
                k = k+2
            
            cpop = np.resize(offspring,[self.numinv,tamgen])
            crpo = np.resize(offena,[self.numinv,nrules])
            
            i = 0
            for mutant in range(self.numinv):
                mutr = 0.1 #0.06
                # mutate an individual with probability mtrate
                if random.random() < self.mtrate:
                    cpop[mutant,:] = self.mutation(cpop[mutant,:],mutr)
                    fitt[i] = 0
                i = i+1
            fit[:] = fitt
            
            # Evaluate population
            nnev = 0
            for k in range(self.numinv):
                if k % self.dshow == 0:
                        fin = time.time()
                        delta = fin-inicio
                        inicio = fin
                        print('Eval individuo '+str(k)+' of '+str(self.numinv)+' in '+str(delta)+ ' s')
                if fit[k,1] == 0:
                    fit[k,0],fiteva,restrain,reseval= self.ffitness(cpop[k,:],crpo[k,:])
                    fit[k,1] = 1
                    nnev = nnev+1
                    if fit[k,0] > bfit:
                        bfit = fit[k,0]
                        bfite = fiteva
                        binv = cpop[k,:]
                        beru = crpo[k,:]
                        data_result = restrain
                        data_evaluation = reseval
                        
            # The population is entirely replaced by the offspring
            popd[:] = cpop
            pope[:] = crpo
            
            mean = sum(fit[:,0]) / self.numinv
            sum2 = sum (x*x for x in fit[:,0])
            std = abs(sum2 / self.numinv - mean**2)**0.5
            
            tem = np.zeros(8)
            tem[0] = g
            tem[1] = nnev
            tem[2] = min(fit[:,0])
            tem[3] = max(fit[:,0])
            tem[4] = mean
            tem[5] = std
            tem[6] = bfite
            tem[7] = self.samp
            data = np.append(data,tem)
        print("------------ End of GFS evolution ------------")
        data = np.reshape (data, [g+1,8])
        np.savetxt('Results/'+self.db+'_'+str(self.inp)+'_GFS_result.csv',data,delimiter=',')
        np.savetxt('Results/'+self.db+'_'+str(self.inp)+'_GFS_genome.csv',binv,delimiter=',')
        np.savetxt('Results/'+self.db+'_'+str(self.inp)+'_GFS_enable.csv',beru,delimiter=',')
        np.savetxt('Results/'+self.db+'_'+str(self.inp)+'_GFS_DT.csv',data_result,delimiter=',')
        np.savetxt('Results/'+self.db+'_'+str(self.inp)+'_GFS_DE.csv',data_evaluation,delimiter=',')
        print("Best fitness value %i" %bfit)
            
# TOOLS FUNCTIONS:

# Exist
def exist(a,b):
    tam=len(b)
    exist= False
    for i in range(tam):
        if b[i] == a:
            exist=True
            break
    return exist

# Dec to bin
def dec_to_bin(x,nb):
    ou=np.zeros(nb)    
    for i in range(nb):    
        r=x%2
        ou[nb-(i+1)]=r
        x=int(x/2)
    return ou
    
# Bin to dec
def bin_to_dec(x,nb):
    ou = 0    
    for i in range(nb):
        ind = nb-(i+1)
        ou = ou + x[ind]*2**i
    return ou

# Bin to dec array
def abin_to_adec(x,nb):
    ou=([])
    num = int (len(x)/nb)
    for i in range(num):
        ori = nb*i
        ou = np.append (ou,bin_to_dec(x[ori:ori+nb],nb))
    return ou
    